""" 
Compare refractive indices to the experimentally measured values.
KBBF (KBe2BO3F2)
"""
from numpy import pi
from ndispers.media.crystals import _KBBF_Li2016

modules = [_KBBF_Li2016]

def sub(module):
    module_name = module.__name__
    print("[[ Module : %s ]]" % module_name)

    kbbf = module.KBBF()

    # From Table 1 in Li et al. 2016, Applied Optics, 55(36), 10423-10426
    # Wavelength (µm)
    wl_list = [
        0.1870, 0.1900, 0.1930, 0.2000, 0.2357, 0.3650, 0.4067, 
        0.4358, 0.4800, 0.5461, 0.5876, 0.6438, 0.7065, 0.8521, 
        1.014, 1.530
    ]
    
    # Experimental values for ordinary ray
    n_o_list = [
        1.59005, 1.58444, 1.57995, 1.56922, 1.52622, 1.49600, 1.49128, 
        1.48844, 1.48538, 1.48208, 1.48048, 1.47875, 1.47727, 1.47456, 
        1.47213, 1.46544
    ]
    
    # Experimental values for extraordinary ray
    n_e_list = [
        1.46299, 1.45991, 1.45718, 1.45132, 1.42962, 1.40641, 1.40336, 
        1.40135, 1.39933, 1.39737, 1.39633, 1.39524, 1.39430, 1.39274, 
        1.39142, 1.38821
    ]

    T_degC = 22.0  # Temperature specified in the paper
    print("T=%.1f degC"  % T_degC)

    print("-"*80)
    print("Wavelength(µm)")
    print("  |  n_o        n_e     |    dn_o/dT       dn_e/dT")
    print("  |  Experimental")
    print("  |  Calculated")
    print("  |  (Cal - Exp)")
    print("-"*80)
    for i, wl in enumerate(wl_list):
        # measured
        n_o_exp = n_o_list[i]
        n_e_exp = n_e_list[i]
        # calculated
        n_o_calc = kbbf.n(wl, 0, T_degC, pol='o')
        n_e_calc = kbbf.n(wl, pi*0.5, T_degC, pol='e')
        dndT_o = kbbf.dndT(wl, 0, T_degC, pol='o')
        dndT_e = kbbf.dndT(wl, 0, T_degC, pol='e')

        print("%.4f" % wl)
        print("  |  {: .5f}   {: .5f}  |".format(n_o_exp, n_e_exp))
        print("  |  {: .5f}   {: .5f}  |  {: .4e}   {: .4e}".format(n_o_calc, n_e_calc, dndT_o, dndT_e))
        print("  | ({: .5f}) ({: .5f}) |".format(n_o_calc - n_o_exp, n_e_calc - n_e_exp))

    print("-"*80)
    
    # Test phase-matching angles for Type I SHG (from Table 2)
    print("\nPhase-matching angles for Type I SHG")
    print("-"*80)
    print("Fundamental Wavelength(µm) | SHG Wavelength(µm) | Exp Angle(deg) | Cal Angle(deg) | Difference")
    print("-"*80)
    
    # Data from Table 2 in the paper
    # Format: [fundamental wavelength (nm), experimental angle (deg)]
    pm_data = [
        [340.0, 70.4],
        [345.0, 68.3],
        [352.4, 65.3],
        [354.7, 64.5],
        [361.2, 62.6],
        [363.4, 61.1],
        [367.4, 60.0],
        [369.5, 61.0],
        [374.3, 59.4],
        [378.6, 57.6],
        [384.7, 56.8],
        [388.4, 55.1],
        [410.0, 51.0],
        [440.0, 46.0],
        [460.0, 44.0],
        [480.0, 41.7],
        [500.0, 39.9],
        [520.0, 36.9],
        [550.0, 34.9],
        [589.0, 32.7],
        [600.0, 32.1],
        [680.0, 27.7],
        [770.0, 25.1],
        [850.0, 23.0],
        [900.0, 22.0],
        [950.0, 21.4],
        [1064, 20.2],
        [1200, 19.7],
        [1300, 19.3],
        [1400, 19.3]
    ]
    
    for data in pm_data:
        fund_wl_nm, exp_angle = data
        fund_wl = fund_wl_nm / 1000  # Convert to µm
        shg_wl_nm = fund_wl_nm / 2   # SHG wavelength is half the fundamental
        
        # Calculate phase-matching angle using the library
        # For Type I SHG (o + o -> e), we use the 'ooe' key
        pm_angles = kbbf.pmAngles_sfg(fund_wl, fund_wl, T_degC, tol_deg=0.005, deg=True)
        cal_angle = pm_angles['ooe']['theta'][0] if pm_angles['ooe']['theta'] else None
        
        if cal_angle is not None:
            diff = cal_angle - exp_angle
            print("{:.1f} | {:.1f} | {:.2f} | {:.2f} | {:.2f}".format(
                fund_wl_nm, shg_wl_nm, exp_angle, cal_angle, diff))
        else:
            print("{:.1f} | {:.1f} | {:.2f} | No solution found | N/A".format(
                fund_wl_nm, shg_wl_nm, exp_angle))
    
    print("-"*80)

def main():
    print("="*80)
    print("Compare refractive indices to the experimentally measured values.")
    print("Ref: %s\n"  % ('Li, R., et al. "Dispersion relations of refractive indices suitable for KBe2BO3F2 crystal deep-ultraviolet applications." Applied Optics, 55(36), 10423-10426 (2016)'))
    for module in modules:
        sub(module)
    print("="*80)

if __name__ == "__main__":
    main()



import sympy

from ndispers._baseclass import T, phi, theta, wl
from ndispers.groups import Uniax_neg_3m
from ndispers.helper import vars2


class SLT(Uniax_neg_3m):
    """
    0.5% MgO-doped stoichiometric Lithium tantalate (Li Ta O_3) crystal

    - Point group : 3m  (C_{3v})
    - Crystal system : Trigonal
    - Dielectic principal axis, z // c-axis (x, y-axes are arbitrary)
    - Negative uniaxial, with optic axis parallel to z-axis
    - Tranparency range : 0.28 µm to 5.6 µm

    Sellmeier equation
    ------------------
    n(wl) = sqrt(a1_i + b1_i * f + (a2_i + b2_i * f)/(wl**2 - (a3_i + b3_i * f)**2) + (a4_i + b4_i * f)/(wl**2 - (a5_i + b5_i * f)**2) - a6_i * wl**2) for i=o,e
    f = (T - T0) * (T + T0 + 2 * 273.16) with T0 = 24.5 degC

    Ref
    ---
    - Sellmeier equation, ratio of d22 vs d33:
    Dolev, I., et al. "Linear and nonlinear optical properties of MgO: LiTaO 3." Applied Physics B 96 (2009): 423-432.
    - Second order nonlinear-optical coefficients, d33, d31:
    Shoji, I., et al. "Absolute scale of second-order nonlinear-optical coefficients," J. Opt. Soc. Am. B 14 (1997): 2268-2294
    
    Note
    ----
    d coefficients are of *congrunet* LT crystal, not of stoichiometric.

    Example
    -------
    >>> bbo = ndispers.media.crystals.BetaBBO_Eimerl1987()
    >>> bbo.n(0.6, 0.5*pi, 25, pol='e') # args: (wl_um, theta_rad, T_degC, pol)
    
    """
    __slots__ = ["_LT__plane", "_LT__theta_rad", "_LT__phi_rad",
                 "_a1_o", "_a2_o", "_a3_o", "_a4_o",  "_a5_o", "_a6_o",
                 "_a1_e", "_a2_e", "_a3_e", "_a4_e",  "_a5_e", "_a6_e",
                 "_b1_o", "_b2_o", "_b3_o", "_b4_o", "_b5_o", 
                 "_b1_e", "_b2_e", "_b3_e", "_b4_e", "_b5_e"]

    def __init__(self):
        super().__init__()
        self._LT__plane = 'arb'
        self._LT__theta_rad = 'var'
        self._LT__phi_rad = 'arb'

        """ Constants of dispersion formula """
        # 0.5% MgO-doped SLT
        ## ordinary ray
        self._a1_o = 4.5082
        self._a2_o = 0.084888
        self._a3_o = 0.19552
        self._a4_o = 1.1570
        self._a5_o = 8.2517
        self._a6_o = 0.0237
        self._b1_o = 2.0704e-8
        self._b2_o = 1.4449e-8
        self._b3_o = 1.5978e-8
        self._b4_o = 4.7686e-6
        self._b5_o = 1.1127e-5
        ## extraordinary ray
        self._a1_e = 4.5615
        self._a2_e = 0.08488
        self._a3_e = 0.1927
        self._a4_e = 5.5832
        self._a5_e = 8.3067
        self._a6_e = 0.021696
        self._b1_e = 4.782e-7
        self._b2_e = 3.0913e-8
        self._b3_e = 2.7326e-8
        self._b4_e = 1.4837e-5
        self._b5_e = 1.3647e-7
        # Second-order nonlinear optical coefficients
        self._d33_1064shg = 13.8 #pm/V
        self._d31_1064shg = 0.85 #pm/V
        self._d22_1064shg = 0.12 * 13.8 #pm/V
    
    @property
    def plane(self):
        return self._LT__plane

    @property
    def theta_rad(self):
        return self._LT__theta_rad

    @property
    def phi_rad(self):
        return self._LT__phi_rad

    @property
    def symbols(self):
        return [wl, theta, phi, T]
    
    @property
    def constants(self):
        print(vars2(self))
    
    def n_o_expr(self):
        """ Sympy expression, dispersion formula for o-wave """
        return sympy.sqrt( self._a1_o + self._b1_o * self.f_expr() + \
            (self._a2_o + self._b2_o * self.f_expr()) / (wl**2 - (self._a3_o + self._b3_o * self.f_expr())**2) + \
                (self._a4_o + self._b4_o * self.f_expr()) / (wl**2 - (self._a5_o + self._b5_o * self.f_expr())**2) - self._a6_o * wl**2 )
                
    def n_e_expr(self):
        """ Sympy expression, dispersion formula for o-wave """
        return sympy.sqrt( self._a1_e + self._b1_e * self.f_expr() + \
            (self._a2_e + self._b2_e * self.f_expr()) / (wl**2 - (self._a3_e + self._b3_e * self.f_expr())**2) + \
                (self._a4_e + self._b4_e * self.f_expr()) / (wl**2 - (self._a5_e + self._b5_e * self.f_expr())**2) - self._a6_e * wl**2 )

    def f_expr(self):
        return (T - 24.5) * (T + 24.5 + 2 * 273.16)

    def n_expr(self, pol):
        """"
        Sympy expression, 
        dispersion formula of a general ray with an angle theta to optic axis. If theta = 0, this expression reduces to 'n_o_expr'.

        n(theta) = n_e / sqrt( sin(theta)**2 + (n_e/n_o)**2 * cos(theta)**2 )
        """
        if pol == 'o':
            return self.n_o_expr()
        elif pol == 'e':
            return self.n_e_expr() / sympy.sqrt( sympy.sin(theta)**2 + (self.n_e_expr()/self.n_o_expr())**2 * sympy.cos(theta)**2 )
        else:
            raise ValueError("pol = '%s' must be 'o' or 'e'" % pol)
    
    def n(self, wl_um, theta_rad, T_degC, pol='e'):
        """
        Refractive index as a function of wavelength, theta or phi angles for each eigen polarization of light.

        input
        -----
        wl_um     :  float or array_like, wavelength in µm
        theta_rad :  float or array_like, 0 to pi radians
        T_degC    :  float or array_like, temperature of crystal in degree C.
        pol       :  {'e'}, optional, polarization of light

        return
        -------
        numpy.array
        
        """
        return super().n(wl_um, theta_rad, 0, T_degC, pol=pol)

    def dn_wl(self, wl_um, theta_rad, T_degC, pol='o'):
        return super().dn_wl(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def d2n_wl(self, wl_um, theta_rad, T_degC, pol='o'):
        return super().d2n_wl(wl_um, theta_rad, 0, T_degC, pol=pol)

    def d3n_wl(self, wl_um, theta_rad, T_degC, pol='o'):
        return super().d3n_wl(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def GD(self, wl_um, theta_rad, T_degC, pol='o'):
        """Group Delay [fs/mm]"""
        return super().GD(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def GV(self, wl_um, theta_rad, T_degC, pol='o'):
        """Group Velocity [µm/fs]"""
        return super().GV(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def ng(self, wl_um, theta_rad, T_degC, pol='o'):
        """Group index, c/Group velocity"""
        return super().ng(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def GVD(self, wl_um, theta_rad, T_degC, pol='o'):
        """Group Delay Dispersion [fs^2/mm]"""
        return super().GVD(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def TOD(self, wl_um, theta_rad, T_degC, pol='o'):
        """Third Order Dispersion [fs^3/mm]"""
        return super().TOD(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def woa_theta(self, wl_um, theta_rad, T_degC, pol='e'):
        """ Polar walk-off angle [rad] """
        return super().woa_theta(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def woa_phi(self, wl_um, theta_rad, T_degC, pol='e'):
        """ Azimuthal walk-off angle [rad] """
        return super().woa_phi(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def dndT(self, wl_um, theta_rad, T_degC, pol='o'):
        return super().dndT(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    #------------------------------------------------------------------------------------------
    # Wavelength dependence of second-order nonlinear coefficients estimated from Miller's rule
    #------------------------------------------------------------------------------------------
    def d22_sfg(self, wl1o, wl2o, T_degC):
        return super().d22_sfg(wl1o, wl2o, T_degC, delta22=self.delta22(self._d22_1064shg, 1.064, 1.064, T_degC))

    def d31_sfg(self, wl1o, wl2o, T_degC):
        return super().d31_sfg(wl1o, wl2o, T_degC, delta31=self.delta31(self._d31_1064shg, 1.064, 1.064, T_degC))


""" 
Compare refractive indices to the experimentaly measured.
LBO
"""
from ndispers.media.crystals import (_LBO_Castech, _LBO_Ghosh1995, _LBO_Newlight, _LBO_KK1994, _LBO_KK2018)
modules = [_LBO_Castech, _LBO_Ghosh1995, _LBO_Newlight, _LBO_KK1994, _LBO_KK2018]

def sub(module):
    module_name = module.__name__
    print("[[ Module : %s ]]" % module_name)

    lbo_yz = module.LBO_yz() # n_x
    lbo_zx = module.LBO_zx() # n_y
    lbo_xy = module.LBO_xy() # n_z

    # from Nikogosyan, Nonlinear Optical Crystals, LBO ref[1]
    wl_list =  [0.2537, 0.3341, 0.5321, 1.0642]
    n_x_list = [1.6335, 1.6043, 1.5785, 1.5656]
    n_y_list = [1.6582, 1.6346, 1.6065, 1.5905]
    n_z_list = [1.6792, 1.6509, 1.6212, 1.6055]

    T_degC = 20.0
    print("T=%.1f degC"  % T_degC)

    print("-"*80)
    print("Wavelength(µm)")
    print("  |  n_x        n_y        n_z    |    dn_x/dT       dn_y/dT       dn_z/dT")
    print("  |  Experimental")
    print("  |  Calculated")
    print("  |  (Cal - Exp)")
    print("-"*80)
    for i, wl in enumerate(wl_list):
        # measured
        n_x_exp = n_x_list[i]
        n_y_exp = n_y_list[i]
        n_z_exp = n_z_list[i]
        # calculated
        n_x_calc = lbo_yz.n(wl, 0, T_degC, pol='o')
        n_y_calc = lbo_zx.n(wl, 0, T_degC, pol='o')
        n_z_calc = lbo_xy.n(wl, 0, T_degC, pol='o')

        dndT_x = lbo_yz.dndT(wl, 0, T_degC, pol='o')
        dndT_y = lbo_zx.dndT(wl, 0, T_degC, pol='o')
        dndT_z = lbo_xy.dndT(wl, 0, T_degC, pol='o')

        print("%.4f" % wl)
        print("  |  {: .4f}   {: .4f}   {: .4f}  |".format(n_x_exp, n_y_exp, n_z_exp))
        print("  |  {: .4f}   {: .4f}   {: .4f}  |  {: .4e}   {: .4e}   {: .4e}".format(n_x_calc, n_y_calc, n_z_calc, dndT_x, dndT_y, dndT_z))
        print("  | ({: .4f}) ({: .4f}) ({: .4f}) |".format(n_x_calc - n_x_exp, n_y_calc - n_y_exp, n_z_calc - n_z_exp))

    print("-"*80)

def main():
    print("="*80)
    print("Compare refractive indices to the experimentaly measured values.")
    print("Ref: %s\n" %('Nikogosyan, D.N., Nonlinear Optical Crystals: A Complete Survey (Springer, 2005)'))
    for module in modules:
        sub(module)

if __name__ == "__main__":
    main()

    


""" 
Compare refractive indices, thermo-optic coefficients, and phase-matching angles to the experimentally measured values.
RBBF (RbBe2BO3F2)
"""
import numpy as np
from numpy import pi

from ndispers.media.crystals import _RBBF_Chen2009

modules = [_RBBF_Chen2009]

def sub(module):
    module_name = module.__name__
    print("[[ Module : %s ]]" % module_name)

    rbbf = module.RBBF()

    # From Table 2 in Chen et al. 2009, JOSAB 26(8), 1519-1525
    # Wavelength (nm)
    wl_list =   [404.7, 435.8, 486.1, 491.6, 546.1, 577.0, 589.3, 656.3, 694.3]
    # Convert to µm
    wl_list = [wl/1000 for wl in wl_list]
    
    # Experimental values for ordinary ray
    n_o_list = [1.49761, 1.49469, 1.49128, 1.49092, 1.48827, 1.48706, 1.48636, 1.48468, 1.48384]
    # Experimental values for extraordinary ray
    n_e_list = [1.41956, 1.41748, 1.41511, 1.41493, 1.41314, 1.41238, 1.41178, 1.41071, 1.41011]

    T_degC = 20.0
    print("T=%.1f degC"  % T_degC)

    print("-"*80)
    print("Wavelength(µm)")
    print("  |  n_o        n_e     |    dn_o/dT       dn_e/dT")
    print("  |  Experimental")
    print("  |  Calculated")
    print("  |  (Cal - Exp)")
    print("-"*80)
    for i, wl in enumerate(wl_list):
        # measured
        n_o_exp = n_o_list[i]
        n_e_exp = n_e_list[i]
        # calculated
        n_o_calc = rbbf.n(wl, 0, T_degC, pol='o')
        n_e_calc = rbbf.n(wl, pi*0.5, T_degC, pol='e')
        dndT_o = rbbf.dndT(wl, 0, T_degC, pol='o')
        dndT_e = rbbf.dndT(wl, 0, T_degC, pol='e')

        print("%.4f" % wl)
        print("  |  {: .5f}   {: .5f}  |".format(n_o_exp, n_e_exp))
        print("  |  {: .5f}   {: .5f}  |  {: .4e}   {: .4e}".format(n_o_calc, n_e_calc, dndT_o, dndT_e))
        print("  | ({: .5f}) ({: .5f}) |".format(n_o_calc - n_o_exp, n_e_calc - n_e_exp))

    print("-"*80)
    
    # Test phase-matching angles for Type I SHG (from Table 3)
    print("\nPhase-matching angles for Type I SHG")
    print("-"*80)
    print("Fundamental Wavelength(µm) | SHG Wavelength(µm) | Exp Angle(deg) | Cal Angle(deg) | Difference")
    print("-"*80)
    
    # Data from Table 3 in the paper
    # Format: [fundamental wavelength (nm), SHG wavelength (nm), experimental angle (deg)]
    pm_data = [
        [354.7, 177.3, 73.38],
        [360.0, 180.0, 70.31],
        [365.0, 182.5, 68.64],
        [370.0, 185.0, 66.92],
        [375.0, 187.5, 65.04],
        [380.0, 190.0, 63.51],
        [385.0, 192.5, 61.85],
        [390.0, 195.0, 60.60],
        [395.0, 197.5, 59.49],
        [400.0, 200.0, 58.04],
        [405.0, 202.5, 56.94],
        [410.0, 205.0, 55.83],
        [415.0, 207.5, 54.88],
        [420.0, 210.0, 54.14],
        [425.0, 212.5, 53.39],
        [430.0, 215.0, 52.37],
        [435.0, 217.5, 51.58],
        [440.0, 220.0, 50.81],
        [515.0, 257.5, 41.17],
        [529.6, 264.8, 39.86],
        [532.0, 266.0, 39.97],
        [549.7, 274.9, 38.24],
        [570.2, 285.1, 36.74],
        [589.7, 294.9, 35.38],
        [610.0, 305.0, 34.00],
        [629.7, 314.9, 32.89],
        [664.5, 332.3, 31.38],
        [730.0, 365.0, 28.55],
        [740.0, 370.0, 28.20],
        [750.0, 375.0, 27.82],
        [750.1, 375.1, 27.83],
        [760.0, 380.0, 27.55],
        [760.8, 380.4, 27.56],
        [770.0, 385.0, 27.16],
        [780.0, 390.0, 26.93],
        [790.0, 395.0, 26.55],
        [799.7, 399.9, 26.30],
        [800.0, 400.0, 26.26],
        [810.0, 405.0, 26.04],
        [812.2, 406.1, 26.12],
        [820.0, 410.0, 25.81],
        [830.0, 415.0, 25.52],
        [840.0, 420.0, 25.28],
        [849.4, 424.7, 25.05],
        [850.0, 425.0, 25.06],
        [860.0, 430.0, 24.80],
        [870.0, 435.0, 24.66],
        [880.0, 440.0, 24.36],
        [897.7, 448.9, 23.93],
        [949.8, 474.9, 23.22],
        [997.7, 498.9, 22.53],
        [1064.0, 532.0, 21.62],
        [1109.0, 554.5, 21.42],
        [1203.1, 601.6, 20.90],
        [1299.2, 649.6, 20.44],
        [1399.5, 699.8, 20.34]
    ]
    
    for data in pm_data:
        fund_wl_nm, shg_wl_nm, exp_angle = data
        fund_wl = fund_wl_nm / 1000  # Convert to µm
        
        # Calculate phase-matching angle using the library
        # For Type I SHG (o + o -> e), we use the 'ooe' key
        pm_angles = rbbf.pmAngles_sfg(fund_wl, fund_wl, T_degC, tol_deg=0.005, deg=True)
        cal_angle = pm_angles['ooe']['theta'][0] if pm_angles['ooe']['theta'] else None
        
        if cal_angle is not None:
            diff = cal_angle - exp_angle
            print("{:.1f} | {:.1f} | {:.2f} | {:.2f} | {:.2f}".format(
                fund_wl_nm, shg_wl_nm, exp_angle, cal_angle, diff))
        else:
            print("{:.1f} | {:.1f} | {:.2f} | No solution found | N/A".format(
                fund_wl_nm, shg_wl_nm, exp_angle))
    
    print("-"*80)

def test_dndT(module):
    """Test thermo-optic coefficients (dn/dT) against experimental values from Zhai et al. 2013"""
    module_name = module.__name__
    print("[[ Module : %s - Thermo-optic Coefficients ]]" % module_name)
    
    rbbf = module.RBBF()
    
    # Data from Table 2 in Zhai et al. 2013, Optical Materials 36(2), 333-336
    # Format: wavelength (µm), experimental dn/dT for o-ray (×10⁻⁶), calculated dn/dT for o-ray (×10⁻⁶),
    #        experimental dn/dT for e-ray (×10⁻⁶), calculated dn/dT for e-ray (×10⁻⁶)
    dndT_data = [
        [0.194, -6.78, -6.78, -7.35, -7.35],
        [0.254, -10.00, -9.99, -9.98, -9.97],
        [0.363, -11.29, -11.36, -9.91, -9.97],
        [0.405, -11.54, -11.54, -9.95, -9.91],
        [0.435, -11.67, -11.63, -9.92, -9.91],
        [0.546, -11.85, -11.84, -10.09, -10.06],
        [0.644, -11.95, -11.96, -10.30, -10.33],
        [0.706, -12.02, -12.03, -10.52, -10.52],
        [1.014, -12.27, -12.27, -11.48, -11.47]
    ]
    
    T_degC = 24.0  # Reference temperature from the paper
    
    print("T=%.1f degC" % T_degC)
    print("-"*100)
    print("Wavelength(µm) | dn_o/dT (×10⁻⁶) |                | dn_e/dT (×10⁻⁶) |               ")
    print("              | Experimental   | Calculated     | Experimental   | Calculated    ")
    print("              |                | (Difference)   |                | (Difference)  ")
    print("-"*100)
    
    for data in dndT_data:
        wl, dndT_o_exp, dndT_o_cal_paper, dndT_e_exp, dndT_e_cal_paper = data
        
        # Calculate dn/dT using our implementation
        dndT_o_calc = rbbf.dndT(wl, 0, T_degC, pol='o') * 1e6  # Convert to ×10⁻⁶
        dndT_e_calc = rbbf.dndT(wl, pi/2, T_degC, pol='e') * 1e6  # Convert to ×10⁻⁶
        
        # Calculate differences
        diff_o = dndT_o_calc - dndT_o_exp
        diff_e = dndT_e_calc - dndT_e_exp
        
        print("{:14.3f} | {:14.2f} | {:14.2f} | {:14.2f} | {:14.2f}".format(
            wl, dndT_o_exp, dndT_o_calc, dndT_e_exp, dndT_e_calc))
        print("              |                | ({:+10.2f})   |                | ({:+10.2f})  ".format(
            diff_o, diff_e))
    
    print("-"*100)
    
    # Calculate average absolute differences
    avg_diff_o = np.mean([abs(data[1] - rbbf.dndT(data[0], 0, T_degC, pol='o') * 1e6) for data in dndT_data])
    avg_diff_e = np.mean([abs(data[3] - rbbf.dndT(data[0], pi/2, T_degC, pol='e') * 1e6) for data in dndT_data])
    
    print("Average absolute difference (o-ray): {:.4f} × 10⁻⁶".format(avg_diff_o))
    print("Average absolute difference (e-ray): {:.4f} × 10⁻⁶".format(avg_diff_e))
    print()

def main():
    print("="*100)
    print("Compare refractive indices, thermo-optic coefficients, and phase-matching angles to the experimentally measured values.")
    print("Refs:\n  - Chen, C., et al. \"Growth, properties, and application to nonlinear optics of a nonlinear optical crystal: RbBe2BO3F2.\" Journal of the Optical Society of America B, 26(8), 1519-1525 (2009)")
    print("  - Zhai, N., et al. \"Measurement of thermal refractive index coefficients of nonlinear optical crystal RbBe2BO3F2.\" Optical Materials, 36(2), 333-336 (2013)")
    print()
    
    for module in modules:
        sub(module)
        test_dndT(module)
    print("="*100)

if __name__ == "__main__":
    main()



from ndispers._baseclass import Medium
import numpy as np

class Uniax_neg_3m(Medium):
    """
    Class of negative uniaxial crystals of point group 3m.

    d = [[   0,   0,   0,   0, d15, -d22 ],
         [ -d22, d22,  0,  d15,  0,   0  ],
         [  d31, d31, d33,  0,   0,   0  ]]
      = [d_il]
        i = 1,2,3 (1=x, 2=y, 3=z)
        l = 1,2,...,6 (1=xx, 2=yy, 3=zz, 4=yz, 5=zx, 6=xy)

    Under the Kleinman symmetry,
    dxzx = dzxx, or d15 = d31.
    Nonzero independent elements are d22(=-d16), d31(=d15) and d33.
    (d33 does not come up in deff since its geometric factor is zero for all the interaction types.)

    """

    def __init__(self):
        super().__init__()

    def delta22(self, d22_known, wl1, wl2, T_degC):
        """
        Miller' delta for d22 coefficient (ijk = yyy).

        Note
        ----
        This value is almost constant and can be used to estimate wavelength dependence of d22
        from a known d22 of SFG with wl1 and wl2 as two pump wavelengths (Miller's rule).
        For yyy, two pump (wl1, wl2) and sum-frequency (wl3) waves are all ordinary waves.

        Parameters
        ----------
        d22_known : float, second-order nonlinear coefficient, d22, known from literature.
        wl1 : float, first pump wavelength in µm for the d22
        wl2 : float, second pump wavelength in µm for the d22
        T_degC : float, crystal temperature in degC.

        return
        ------
        float, Miller's delta for 22 (ijk = yyy) component.

        """
        wl3 = 1./(1./wl1 + 1./wl2)
        theta_rad = 0 # This is an arbitrary value since n of o-wave does not depends on theta.
        # chi : linear susceptibility
        chi_wl1o = self.n(wl1, theta_rad, T_degC, pol='o')**2 - 1
        chi_wl2o = self.n(wl2, theta_rad, T_degC, pol='o')**2 - 1
        chi_wl3o = self.n(wl3, theta_rad, T_degC, pol='o')**2 - 1
        return d22_known / (chi_wl3o * chi_wl2o * chi_wl1o)

    def d22_sfg(self, wl1, wl2, T_degC, delta22=0):
        """
        d22 coefficient as a function of two pump wavelengths for SFG.

        Parameters
        ----------
        wl1 : float, first pump wavelength in µm.
        wl2 : float, second pump wavelength in µm.
        T_degC : crystal temperature in degC.

        return
        ------
        float, d22 coefficient (pm/V).

        """

        wl3 = 1./(1./wl1 + 1./wl2)
        theta_rad = 0 # This is an arbitrary value since n of o-wave does not depends on theta.
        chi_wl1o = self.n(wl1, theta_rad, T_degC, pol='o')**2 - 1
        chi_wl2o = self.n(wl2, theta_rad, T_degC, pol='o')**2 - 1
        chi_wl3o = self.n(wl3, theta_rad, T_degC, pol='o')**2 - 1
        return delta22 * chi_wl3o * chi_wl2o * chi_wl1o
    
    def delta31(self, d31_known, wl1, wl2, T_degC):
        """
        Miller' delta for d31 coefficient (ijk = zxx).

        Note
        ----
        This value is almost constant and can be used to estimate wavelength dependence of d31
        from a known d31 of SFG with wl1 and wl2 as two pump wavelengths (Miller's rule).
        For zxx, two pump waves are ordinary and the SF wave is extra-ordinary.

        Parameters
        ----------
        d31_known : float, second-order nonlinear coefficient, d31, known from literature.
        wl1 : float, first pump wavelength in µm for the d31
        wl2 : float, second pump wavelength in µm for the d31
        T_degC : float, crystal temperature in degC.

        return
        ------
        float, Miller's delta for 31 (ijk = zxx) component.

        """
        wl3 = 1./(1./wl1 + 1./wl2)
        pmAngles = self.pmAngles_sfg(wl1, wl2, T_degC)
        theta_rad = pmAngles['ooe']['theta'][0] # rad
        chi_wl1o = self.n(wl1, theta_rad, T_degC, pol='o')**2 - 1
        chi_wl2o = self.n(wl2, theta_rad, T_degC, pol='o')**2 - 1
        chi_wl3e = self.n(wl3, theta_rad, T_degC, pol='e')**2 - 1
        return d31_known / (chi_wl3e * chi_wl2o * chi_wl1o)
    
    def d31_sfg(self, wl1, wl2, T_degC, delta31=0):
        """
        d31 coefficient as a function of two pump wavelengths for SFG.

        Parameters
        ----------
        wl1 : float, first pump wavelength in µm.
        wl2 : float, second pump wavelength in µm.
        T_degC : float, crystal temperature in degC.

        return
        ------
        float, d31 coefficient (pm/V).

        """

        wl3 = 1./(1./wl1 + 1./wl2)
        pmAngles = self.pmAngles_sfg(wl1, wl2, T_degC)
        theta_rad = pmAngles['ooe']['theta'][0] # rad
        chi_wl1o = self.n(wl1, theta_rad, T_degC, pol='o')**2 - 1
        chi_wl2o = self.n(wl2, theta_rad, T_degC, pol='o')**2 - 1
        chi_wl3e = self.n(wl3, theta_rad, T_degC, pol='e')**2 - 1
        return delta31 * chi_wl3e * chi_wl2o * chi_wl1o
    
    def deff_sfg(self, wl1, wl2, theta_rad, phi_rad, T_degC, pol1, pol2, pol3):
        """
        Effective second-order nonlinear coefficient, d_eff, of SFG.
        Wavelength dependence is estimated from Miller's rule.

        Parameters
        ----------
        wl1  :  float, first pump wavelength in µm.
        wl2  : float, second pump wavelength in µm.
        T_degC : float, crystal temperature in degC.
        pol1 : str, {'o', 'e'}, polarization of the 1st pump wave.
        pol2 : str, {'o', 'e'}, polarization of the 2nd pump wave.
        pol3 : str, {'o', 'e'}, polarization of the sum-frequency wave.

        return
        ------
        float, Effective d coefficient (pm/V).

        """
        wl3 = 1./(1./wl1 + 1./wl2)
        _d31 = self.d31_sfg(wl1, wl2, T_degC)
        _d22 = self.d22_sfg(wl1, wl2, T_degC)
        rho3 = self.woa_theta(wl3, theta_rad, T_degC, pol=pol3)
        # For negative uniaxial type-I (ooe),
        if pol1 == 'o' and pol2 == 'o' and pol3 == 'e':
            deff = _d31 * np.sin(theta_rad + rho3) - _d22 * np.cos(theta_rad + rho3) * np.sin(3*phi_rad)
        
        # For negative uniaxial type-II (oee),
        elif pol1 == 'o' and pol2 == 'e' and pol3 == 'e':
            rho2 = self.woa_theta(wl2, theta_rad, T_degC, pol=pol2)
            deff = _d22 * np.cos(theta_rad + rho2) * np.cos(theta_rad + rho3) * np.cos(3*phi_rad)
        
        # For negative uniaxial type-II (eoe),
        elif pol1 == 'e' and pol2 == 'o' and pol3 == 'e':
            rho1 = self.woa_theta(wl1, theta_rad, T_degC, pol=pol1)
            deff = _d22 * np.cos(theta_rad + rho1) * np.cos(theta_rad + rho3) * np.cos(3*phi_rad)
        
        else:
            raise ValueError("For negative uniaxial crystals,pol1, pol2, pol3 must be among {ooe, oee, eoe}.")

        return deff



import sympy
from ndispers._baseclass import wl, phi, theta, T
from ndispers.groups import Uniax_neg_3m
from ndispers.helper import vars2

class BetaBBO(Uniax_neg_3m):
    """
    β-BBO (β-Ba B_2 O_4) crystal

    - Point group : 3m  (C_{3v})
    - Crystal system : Trigonal
    - Dielectic principal axis, z // c-axis (x, y-axes are arbitrary)
    - Negative uniaxial, with optic axis parallel to z-axis
    - Tranparency range : 0.19 µm to 2.6 µm

    Sellmeier equation
    ------------------
    n(wl) = sqrt(A_i + B_i/(1 - C_i/wl**2) + D_i/(1 - E_i/wl**2)) + dn/dT * (T -20)

    Thermo-optic coefficient
    ------------------------
    dn/dT = (G_i * R_i + H_i * R_i**2) / 2*n_i  for i = o, e
    R_i = wl**2/(wl**2 - wl0_i**2), where wl0_i = 0.0652 for i = o or wl0 = 0.0730  for i = e

    Ref
    ----
    Sellmeier equation:
    Ghosh, Gorachand. "Temperature dispersion of refractive indices in β-BaB2O4 and LiB3O5 crystals for nonlinear optical devices." Journal of applied physics 78.11 (1995): 6752-6760.
    
    Nonlinear optical coefficients:
    - Shoji, Ichiro, et al. "Absolute measurement of second-order nonlinear-optical coefficients of β-BaB2O4 for visible to ultraviolet second-harmonic wavelengths." JOSA B 16.4 (1999): 620-624.

    Example
    -------
    >>> bbo = ndispers.media.crystals.BetaBBO_Eimerl1987()
    >>> bbo.n(0.6, 0.5*pi, 25, pol='e') # args: (wl_um, theta_rad, T_degC, pol)
    
    """
    __slots__ = ["_BetaBBO__plane", "_BetaBBO__theta_rad", "_BetaBBO__phi_rad",
                 "_A_o", "_B_o", "_C_o", "_D_o", "_E_o",
                 "_A_e", "_B_e", "_C_e", "_D_e", "_E_e",
                 "_G_o", "_H_o", "_R_o",
                 "_G_e", "_H_e", "_R_e",
                 "_d31_1064shg", "_d22_1064shg"]
                 
    def __init__(self):
        super().__init__()
        self._BetaBBO__plane = 'arb'
        self._BetaBBO__theta_rad = 'var'
        self._BetaBBO__phi_rad = 'arb'

        """ Constants of dispersion formula """
        # For ordinary ray
        self._A_o = 1.7018379
        self._B_o = 1.0357554
        self._C_o = 0.018003440
        self._D_o = 1.2479989
        self._E_o = 91
        # For extraordinary ray
        self._A_e = 1.5920433
        self._B_e = 0.7816893
        self._C_e = 0.016067891
        self._D_e = 0.8403893
        self._E_e = 91
        # dn/dT
        self._G_o = -19.3007e-6
        self._G_e = -141.421e-6
        self._H_o = -34.9683e-6
        self._H_e = 110.8630e-6
        self._R_o = wl**2/(wl**2 - 0.0652**2)
        self._R_e = wl**2/(wl**2 - 0.0730**2)
        # Second-order nonlinear optical coefficients
        self._d31_1064shg = 0.04 #pm/V
        self._d22_1064shg = 2.2 #pm/V
    
    @property
    def plane(self):
        return self._BetaBBO__plane

    @property
    def theta_rad(self):
        return self._BetaBBO__theta_rad
        
    @property
    def phi_rad(self):
        return self._BetaBBO__phi_rad
    
    @property
    def symbols(self):
        return [wl, theta, phi, T]
    
    @property
    def constants(self):
        print(vars2(self))
    
    def _n_T20_o_expr(self):
        """ Sympy expression, dispersion formula for o-wave at 20degC """
        return sympy.sqrt(self._A_o + self._B_o / (1 - self._C_o/wl**2) + self._D_o/(1 - self._E_o/wl**2))
    
    def _n_T20_e_expr(self):
        """ Sympy expression, dispersion formula for theta=90 deg e-wave at 20degC """
        return sympy.sqrt(self._A_e + self._B_e / (1 - self._C_e/wl**2) + self._D_e/(1 - self._E_e/wl**2))
    
    def dndT_o_expr(self):
        return (self._G_o * self._R_o + self._H_o * self._R_o**2) / (2*self._n_T20_o_expr())
    
    def dndT_e_expr(self):
        return (self._G_e * self._R_e + self._H_e * self._R_e**2) / (2*self._n_T20_e_expr())
    
    def n_o_expr(self):
        return self._n_T20_o_expr() + self.dndT_o_expr() * (T - 20)
    
    def n_e_expr(self):
        return self._n_T20_e_expr() + self.dndT_e_expr() * (T - 20)

    def n_expr(self, pol):
        """"
        Sympy expression, 
        dispersion formula of a general ray with an angle theta to optic axis. If theta = 0, this expression reduces to 'no_expre'.

        n(theta) = n_e / sqrt( sin(theta)**2 + (n_e/n_o)**2 * cos(theta)**2 )
        
        """
        if pol == 'o':
            return self.n_o_expr()
        elif pol == 'e':
            return self.n_e_expr() / sympy.sqrt( sympy.sin(theta)**2 + (self.n_e_expr()/self.n_o_expr())**2 * sympy.cos(theta)**2 )
        else:
            raise ValueError("pol = '%s' must be 'o' or 'e'" % pol)
    
    def n(self, wl_um, theta_rad, T_degC, pol='o'):
        """
        Refractive index as a function of wavelength, theta and phi angles for each eigen polarization of light.

        input
        ------
        wl_um     :  float or array_like, wavelength in µm
        theta_rad :  float or array_like, 0 to pi radians
        T_degC    :  float or array_like, temperature of crystal in degree C.
        pol       :  {'o', 'e'}, optional, polarization of light

        return
        -------
        Refractive index, float or array_like

        """
        return super().n(wl_um, theta_rad, 0, T_degC, pol=pol)

    def dn_wl(self, wl_um, theta_rad, T_degC, pol='o'):
        return super().dn_wl(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def d2n_wl(self, wl_um, theta_rad, T_degC, pol='o'):
        return super().d2n_wl(wl_um, theta_rad, 0, T_degC, pol=pol)

    def d3n_wl(self, wl_um, theta_rad, T_degC, pol='o'):
        return super().d3n_wl(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def GD(self, wl_um, theta_rad, T_degC, pol='o'):
        """Group Delay [fs/mm]"""
        return super().GD(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def GV(self, wl_um, theta_rad, T_degC, pol='o'):
        """Group Velocity [µm/fs]"""
        return super().GV(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def ng(self, wl_um, theta_rad, T_degC, pol='o'):
        """Group index, c/Group velocity"""
        return super().ng(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def GVD(self, wl_um, theta_rad, T_degC, pol='o'):
        """Group Delay Dispersion [fs^2/mm]"""
        return super().GVD(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def TOD(self, wl_um, theta_rad, T_degC, pol='o'):
        """Third Order Dispersion [fs^3/mm]"""
        return super().TOD(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def woa_theta(self, wl_um, theta_rad, T_degC, pol='e'):
        """ Polar walk-off angle [rad] """
        return super().woa_theta(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def woa_phi(self, wl_um, theta_rad, T_degC, pol='e'):
        """ Azimuthal walk-off angle [rad] """
        return super().woa_phi(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def dndT(self, wl_um, theta_rad, T_degC, pol='o'):
        return super().dndT(wl_um, theta_rad, 0, T_degC, pol=pol)

    #------------------------------------------------------------------------------------------
    # Wavelength dependence of second-order nonlinear coefficients estimated from Miller's rule
    #------------------------------------------------------------------------------------------
    def d22_sfg(self, wl1o, wl2o, T_degC):
        return super().d22_sfg(wl1o, wl2o, T_degC, delta22=self.delta22(self._d22_1064shg, 1.064, 1.064, T_degC))

    def d31_sfg(self, wl1o, wl2o, T_degC):
        return super().d31_sfg(wl1o, wl2o, T_degC, delta31=self.delta31(self._d31_1064shg, 1.064, 1.064, T_degC))


import sympy

from ndispers._baseclass import Medium, T, phi, theta, wl
from ndispers.helper import vars2


class RBBF(Medium):
    """
    RBBF (RbBe2BO3F2, Rubidium Beryllium Borate Fluoride) crystal
    
    - Point group : 32  (D_3)
    - Crystal system : Trigonal
    - Dielectic principal axis, z // c-axis (x, y-axes are arbitrary)
    - Negative uniaxial, with optic axis parallel to z-axis
    - Tranparency range : 0.16 to 3.55 µm
    - Hardness: 2.9 on the Mohs scale
    - Highly stable in air and even in hot water at 100°C or in acids such as HNO3 and HCl

    Sellmeier equation
    ------------------
    n_o^2 = 1 + 1.18675λ²/(λ² - 0.00750) - 0.00910λ²  (λ is in µm)
    n_e^2 = 1 + 0.97530λ²/(λ² - 0.00665) - 0.00145λ²  (λ is in µm)
    
    Thermo-optic dispersion equation
    ------------------------------
    dn/dT = (A/λ³ + B/λ² + C/λ + D) × 10⁻⁶  (λ is in µm)
    
    For ordinary ray (n_o):
    A = 0.099911, B = -0.553474, C = 1.454609, D = -13.260115
    
    For extraordinary ray (n_e):
    A = 0.285633, B = -2.482927, C = 6.916728, D = -16.153736
    
    Validity range
    ---------------
    Sellmeier equation: Deep UV to near infrared
    Thermo-optic equation: 0.194µm ≤ λ ≤ 1.014µm
    
    Implementation notes
    ------------------
    RBBF is a uniaxial crystal where the refractive index doesn't depend on the azimuthal angle phi.
    In all method calls to the parent class, a value of 0 is passed for the phi_rad parameter.

    Ref
    ----
    - Chen, C., Wu, Y., Li, Y., Wang, J., Wu, B., Jiang, M., Zhang, G., & Ye, N. (2009). Growth, properties, and application to nonlinear optics of a nonlinear optical crystal: RbBe2BO3F2. Journal of the Optical Society of America B, 26(8), 1519-1525. https://opg.optica.org/josab/abstract.cfm?uri=josab-26-8-1519
    - Zhai, N., Wang, L., Liu, L., Wang, X., Zhu, Y., & Chen, C. (2013). Measurement of thermal refractive index coefficients of nonlinear optical crystal RbBe2BO3F2. Optical Materials, 36(2), 333-336.

    Example
    -------
    >>> rbbf = ndispers.media.crystals.RBBF()
    >>> rbbf.n(0.6, 0.5*pi, 25, pol='e') # args: (wl_um, theta_rad, T_degC, pol)
    
    """
    __slots__ = ["_RBBF__plane", "_RBBF__theta_rad", "_RBBF__phi_rad",
                 "_B_o", "_C_o", "_D_o", 
                 "_B_e", "_C_e", "_D_e",
                 "_dndT_o_A", "_dndT_o_B", "_dndT_o_C", "_dndT_o_D",
                 "_dndT_e_A", "_dndT_e_B", "_dndT_e_C", "_dndT_e_D"]

    def __init__(self):
        super().__init__()
        self._RBBF__plane = 'arb'
        self._RBBF__theta_rad = 'var'
        self._RBBF__phi_rad = 'arb'

        """ Constants of dispersion formula """
        # For ordinary ray (from the paper's equation: n_o^2 = 1 + 1.18675λ²/(λ² - 0.00750) - 0.00910λ²)
        self._B_o = 1.18675  # Numerator of first term
        self._C_o = 0.00750  # Denominator constant of first term
        self._D_o = 0.00910  # Coefficient of λ² term
        
        # For extraordinary ray (from the paper's equation: n_e^2 = 1 + 0.97530λ²/(λ² - 0.00665) - 0.00145λ²)
        self._B_e = 0.97530  # Numerator of first term
        self._C_e = 0.00665  # Denominator constant of first term
        self._D_e = 0.00145  # Coefficient of λ² term
        
        # dn/dT coefficients from Zhai et al. (2013)
        # Formula: dn/dT = (A/λ³ + B/λ² + C/λ + D) × 10⁻⁶
        # Valid for 0.194μm ≤ λ ≤ 1.014μm
        # Ordinary ray
        self._dndT_o_A = 0.099911  # coefficient of 1/λ³
        self._dndT_o_B = -0.553474  # coefficient of 1/λ²
        self._dndT_o_C = 1.454609  # coefficient of 1/λ
        self._dndT_o_D = -13.260115  # coefficient of constant term
        
        # Extraordinary ray
        self._dndT_e_A = 0.285633  # coefficient of 1/λ³
        self._dndT_e_B = -2.482927  # coefficient of 1/λ²
        self._dndT_e_C = 6.916728  # coefficient of 1/λ
        self._dndT_e_D = -16.153736  # coefficient of constant term
    
    @property
    def plane(self):
        return self._RBBF__plane

    @property
    def theta_rad(self):
        return self._RBBF__theta_rad

    @property
    def phi_rad(self):
        return self._RBBF__phi_rad

    @property
    def constants(self):
        print(vars2(self))
    
    @property
    def symbols(self):
        return [wl, theta, phi, T]

    def dndT_o_expr(self):
        """ Sympy expression for thermo-optic coefficient of o-wave (dn/dT) """
        # Formula: dn/dT = (A/λ³ + B/λ² + C/λ + D) × 10⁻⁶
        return (self._dndT_o_A / wl**3 + self._dndT_o_B / wl**2 + self._dndT_o_C / wl + self._dndT_o_D) * 1e-6
    
    def n_o_expr(self):
        """ Sympy expression, dispersion formula for o-wave """
        n_o = sympy.sqrt(1 + self._B_o * wl**2 / (wl**2 - self._C_o) - self._D_o * wl**2)
        # Add temperature dependence
        return n_o + self.dndT_o_expr() * (T - 24)  # T_ref = 24°C in the paper
    
    def dndT_e_expr(self):
        """ Sympy expression for thermo-optic coefficient of e-wave (dn/dT) """
        # Formula: dn/dT = (A/λ³ + B/λ² + C/λ + D) × 10⁻⁶
        return (self._dndT_e_A / wl**3 + self._dndT_e_B / wl**2 + self._dndT_e_C / wl + self._dndT_e_D) * 1e-6
        
    def n_e_expr(self):
        """ Sympy expression, dispersion formula for theta=90 deg e-wave """
        n_e = sympy.sqrt(1 + self._B_e * wl**2 / (wl**2 - self._C_e) - self._D_e * wl**2)
        # Add temperature dependence
        return n_e + self.dndT_e_expr() * (T - 24)  # T_ref = 24°C in the paper

    def n_expr(self, pol):
        """
        Sympy expression, 
        dispersion formula of a general ray with an angle theta to optic axis. If theta = 0, this expression reduces to 'no_expre'.

        n(theta) = n_e / sqrt( sin(theta)**2 + (n_e/n_o)**2 * cos(theta)**2 )
        
        """
        if pol == 'o':
            return self.n_o_expr()
        elif pol == 'e':
            return self.n_e_expr() / sympy.sqrt( sympy.sin(theta)**2 + (self.n_e_expr()/self.n_o_expr())**2 * sympy.cos(theta)**2 )
        else:
            raise ValueError("pol = '%s' must be 'o' or 'e'" % pol)
    
    def n(self, wl_um, theta_rad, T_degC, pol='o'):
        """
        Refractive index as a function of wavelength, theta and phi angles for each eigen polarization of light.

        input
        ------
        wl_um     :  float or array_like, wavelength in µm
        theta_rad :  float or array_like, 0 to pi radians
        T_degC    :  float or array_like, temperature of crystal in degree C.
        pol       :  {'o', 'e'}, optional, polarization of light

        return
        -------
        Refractive index, float or array_like
        
        Notes
        -----
        When calling the parent class method, we pass 0 for phi_rad since RBBF is a uniaxial crystal
        where the refractive index doesn't depend on the azimuthal angle phi.
        The argument order in the parent class is (wl_um, theta_rad, phi_rad, T_degC, pol).
        """
        return super().n(wl_um, theta_rad, 0, T_degC, pol=pol)

    def dn_wl(self, wl_um, theta_rad, T_degC, pol='o'):
        return super().dn_wl(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def d2n_wl(self, wl_um, theta_rad, T_degC, pol='o'):
        return super().d2n_wl(wl_um, theta_rad, 0, T_degC, pol=pol)

    def d3n_wl(self, wl_um, theta_rad, T_degC, pol='o'):
        return super().d3n_wl(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def GD(self, wl_um, theta_rad, T_degC, pol='o'):
        """Group Delay [fs/mm]"""
        return super().GD(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def GV(self, wl_um, theta_rad, T_degC, pol='o'):
        """Group Velocity [µm/fs]"""
        return super().GV(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def ng(self, wl_um, theta_rad, T_degC, pol='o'):
        """Group index, c/Group velocity"""
        return super().ng(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def GVD(self, wl_um, theta_rad, T_degC, pol='o'):
        """Group Delay Dispersion [fs^2/mm]"""
        return super().GVD(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def TOD(self, wl_um, theta_rad, T_degC, pol='o'):
        """Third Order Dispersion [fs^3/mm]"""
        return super().TOD(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def woa_theta(self, wl_um, theta_rad, T_degC, pol='e'):
        return super().woa_theta(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def woa_phi(self, wl_um, theta_rad, T_degC, pol='e'):
        return super().woa_phi(wl_um, theta_rad, 0, T_degC, pol=pol)

    def dndT_expr(self, pol):
        """ Sympy expression for dn/dT """
        # Override the base class method to use our custom formulas
        if pol == 'o':
            return self.dndT_o_expr()
        elif pol == 'e':
            return self.dndT_e_expr()
        else:
            raise ValueError("pol = '%s' must be 'o' or 'e'" % pol)
    
    def dndT(self, wl_um, theta_rad, T_degC, pol='o'):
        """Thermo-optic coefficient (dn/dT) [1/°C]
        
        Formula: dn/dT = (A/λ³ + B/λ² + C/λ + D) × 10⁻⁶
        Valid for 0.194μm ≤ λ ≤ 1.014μm
        
        input
        ------
        wl_um     :  float or array_like, wavelength in µm
        theta_rad :  float or array_like, 0 to pi radians
        T_degC    :  float or array_like, temperature of crystal in degree C.
        pol       :  {'o', 'e'}, optional, polarization of light
        
        return
        -------
        dn/dT in 1/°C, float or array_like
        """
        return super().dndT(wl_um, theta_rad, 0, T_degC, pol=pol)



import sympy
from ndispers._baseclass import wl, phi, theta, T
from ndispers.groups import Uniax_neg_32
from ndispers.helper import vars2

class KBBF(Uniax_neg_32):
    """
    KBBF (K Be_2 B O_3 F_2) crystal

    - Point group : 32  (D_2)
    - Crystal system : Trigonal
    - Dielectic principal axis, z // c-axis (x, y-axes are arbitrary)
    - Negative uniaxial, with optic axis parallel to z-axis
    - Tranparency range : 0.155 µm to 3.66 µm

    Sellmeier equation
    ------------------
    n(wl) = sqrt(A_i + B1_i * wl**2 / (wl**2 - C1_i**2) + B2_i * wl**2 / (wl**2 - C2_i**2) - D_i * wl**2)  + dndT_i * (T - 22) for i = o, e

    Ref
    ---
    Sellmeier equation:
    Li, R., Wang, L., Wang, X., Wang, G., & Chen, C. (2016). Dispersion relations of refractive indices suitable for KBe 2 BO 3 F 2 crystal deep-ultraviolet applications. Applied Optics, 55(36), 10423-10426.
    
    Nonlinear optical coefficients:
    Chen, C. T., Wang, G. L., Wang, X. Y., & Xu, Z. Y. (2009). Deep-UV nonlinear optical crystal KBe2BO3F2—discovery, growth, optical properties and applications. Applied Physics B, 97(1), 9-25.
    
    Example
    -------
    >>> bbo = ndispers.media.crystals.BetaBBO_Eimerl1987()
    >>> bbo.n(0.6, 0.5*pi, 25, pol='e') # args: (wl_um, theta_rad, T_degC, pol)
    
    """
    __slots__ = ["_KBBF__plane", "_KBBF__theta_rad", "_KBBF__phi_rad",
                 "_A_o", "_B1_o", "_B2_o", "_C1_o", "_C2_o",  "_D_o",
                 "_A_e", "_B1_e", "_B2_e", "_C1_e", "_C2_e",  "_D_e",
                 "_d11_1064shg", "_d14_1064shg"]
                 
    def __init__(self):
        super().__init__()
        self._KBBF__plane = 'arb'
        self._KBBF__theta_rad = 'var'
        self._KBBF__phi_rad = 'arb'

        """ Constants of dispersion formula """
        # For ordinary ray
        self._A_o =  1.024248
        self._B1_o = 0.9502782
        self._B2_o = 0.1960298
        self._C1_o = 0.0738546
        self._C2_o = 0.1298386
        self._D_o =  0.0113908
        # For extraordinary ray
        self._A_e =  0.9411543
        self._B1_e = 0.8684699
        self._B2_e = 0.1256642
        self._C1_e = 0.0646955
        self._C2_e = 0.1196215
        self._D_e =  0.0044736
        # dn/dT
        # (not known yet)
        self._dndT_o = 0.0 #/degC
        self._dndT_e = 0.0 #/degC
        # Second-order nonlinear optical coefficients
        self._d11_1064shg = 0.47 #pm/V
        self._d14_1064shg = 0.0 #pm/V
    
    @property
    def plane(self):
        return self._KBBF__plane

    @property
    def theta_rad(self):
        return self._KBBF__theta_rad

    @property
    def phi_rad(self):
        return self._KBBF__phi_rad

    @property
    def symbols(self):
        return [wl, theta, phi, T]
    
    @property
    def constants(self):
        print(vars2(self))
    
    def n_o_expr(self):
        """ Sympy expression, dispersion formula for o-wave """
        return sympy.sqrt(self._A_o + self._B1_o * wl**2 / (wl**2 - self._C1_o**2) + self._B2_o * wl**2 / (wl**2 - self._C2_o**2) - self._D_o * wl**2) + self._dndT_o * (T - 22)
    
    def n_e_expr(self):
        """ Sympy expression, dispersion formula for theta=90 deg e-wave """
        return sympy.sqrt(self._A_e + self._B1_e * wl**2 / (wl**2 - self._C1_e**2) + self._B2_e * wl**2 / (wl**2 - self._C2_e**2) - self._D_e * wl**2) + self._dndT_e * (T - 22)

    def n_expr(self, pol):
        """"
        Sympy expression, 
        dispersion formula of a general ray with an angle theta to optic axis. If theta = 0, this expression reduces to 'n_o_expr'.

        n(theta) = n_e / sqrt( sin(theta)**2 + (n_e/n_o)**2 * cos(theta)**2 )
        """
        if pol == 'o':
            return self.n_o_expr()
        elif pol == 'e':
            return self.n_e_expr() / sympy.sqrt( sympy.sin(theta)**2 + (self.n_e_expr()/self.n_o_expr())**2 * sympy.cos(theta)**2 )
        else:
            raise ValueError("pol = '%s' must be 'o' or 'e'" % pol)
    
    def n(self, wl_um, theta_rad, T_degC, pol='o'):
        """
        Refractive index as a function of wavelength, theta or phi angles for each eigen polarization of light.

        input
        -----
        wl_um     :  float or array_like, wavelength in µm
        theta_rad :  float or array_like, 0 to pi radians
        T_degC    :  float or array_like, temperature of crystal in degree C.
        pol       :  {'o', 'e'}, optional, polarization of light

        return
        -------
        numpy.array
        
        """
        return super().n(wl_um, theta_rad, 0, T_degC, pol=pol)

    def dn_wl(self, wl_um, theta_rad, T_degC, pol='o'):
        return super().dn_wl(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def d2n_wl(self, wl_um, theta_rad, T_degC, pol='o'):
        return super().d2n_wl(wl_um, theta_rad, 0, T_degC, pol=pol)

    def d3n_wl(self, wl_um, theta_rad, T_degC, pol='o'):
        return super().d3n_wl(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def GD(self, wl_um, theta_rad, T_degC, pol='o'):
        """Group Delay [fs/mm]"""
        return super().GD(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def GV(self, wl_um, theta_rad, T_degC, pol='o'):
        """Group Velocity [µm/fs]"""
        return super().GV(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def ng(self, wl_um, theta_rad, T_degC, pol='o'):
        """Group index, c/Group velocity"""
        return super().ng(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def GVD(self, wl_um, theta_rad, T_degC, pol='o'):
        """Group Delay Dispersion [fs^2/mm]"""
        return super().GVD(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def TOD(self, wl_um, theta_rad, T_degC, pol='o'):
        """Third Order Dispersion [fs^3/mm]"""
        return super().TOD(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def woa_theta(self, wl_um, theta_rad, T_degC, pol='e'):
        """ Polar walk-off angle [rad] """
        return super().woa_theta(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def woa_phi(self, wl_um, theta_rad, T_degC, pol='e'):
        """ Azimuthal walk-off angle [rad] """
        return super().woa_phi(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    def dndT(self, wl_um, theta_rad, T_degC, pol='o'):
        return super().dndT(wl_um, theta_rad, 0, T_degC, pol=pol)
    
    #------------------------------------------------------------------------------------------
    # Wavelength dependence of second-order nonlinear coefficients estimated from Miller's rule
    #------------------------------------------------------------------------------------------
    def d11_sfg(self, wl1o, wl2o, T_degC):
        return super().d11_sfg(wl1o, wl2o, T_degC, delta11=self.delta11(self._d11_1064shg, 1.064, 1.064, T_degC))

    def d14_sfg(self, wl1o, wl2o, T_degC):
        return super().d14_sfg(wl1o, wl2o, T_degC, delta14=self.delta14(self._d14_1064shg, 1.064, 1.064, T_degC))
    


""" 
Compare refractive indices to the experimentaly measured.
β-BBO
"""
from ndispers.media.crystals import (_betaBBO_Eimerl1987, _betaBBO_Ghosh1995, _betaBBO_KK2010, _betaBBO_Tamosauskas2018)
modules = [_betaBBO_Eimerl1987, _betaBBO_Ghosh1995, _betaBBO_KK2010, _betaBBO_Tamosauskas2018]
from numpy import pi

def sub(module):
    module_name = module.__name__
    print("[[ Module : %s ]]" % module_name)

    bbo = module.BetaBBO()

    # from Nikogosyan, D. N. "Beta barium borate (BBO)." Applied Physics A 52.6 (1991): 359-368.
    wl_list =   [0.2128, 0.2660, 0.3541, 0.5321, 1.0642]
    n_o_list = [1.84707, 1.75707, 1.70556, 1.67493, 1.65510]
    n_e_list = [1.67467, 1.61461, 1.57757, 1.55552, 1.54254]

    T_degC = 20.0
    print("T=%.1f degC"  % T_degC)

    print("-"*80)
    print("Wavelength(µm)")
    print("  |  n_o        n_e     |    dn_o/dT       dn_e/dT")
    print("  |  Experimental")
    print("  |  Calculated")
    print("  |  (Cal - Exp)")
    print("-"*80)
    for i, wl in enumerate(wl_list):
        # measured
        n_o_exp = n_o_list[i]
        n_e_exp = n_e_list[i]
        # calculated
        n_o_calc = bbo.n(wl, 0, T_degC, pol='o')
        n_e_calc = bbo.n(wl, pi*0.5, T_degC, pol='e')
        dndT_o = bbo.dndT(wl, 0, T_degC, pol='o')
        dndT_e = bbo.dndT(wl, 0, T_degC, pol='e')

        print("%.4f" % wl)
        print("  |  {: .4f}   {: .4f}  |".format(n_o_exp, n_e_exp))
        print("  |  {: .4f}   {: .4f}  |  {: .4e}   {: .4e}".format(n_o_calc, n_e_calc, dndT_o, dndT_e))
        print("  | ({: .4f}) ({: .4f}) |".format(n_o_calc - n_o_exp, n_e_calc - n_e_exp))

    print("-"*80)

def main():
    print("="*80)
    print("Compare refractive indices to the experimentaly measured values.")
    print("Ref: %s\n"  % ('Nikogosyan, D. N. "Beta barium borate (BBO)." Applied Physics A 52.6 (1991): 359-368.'))
    for module in modules:
        sub(module)

if __name__ == "__main__":
    main()

    


import sympy
from ndispers._baseclass import Medium, wl, phi, theta, T, pi
from ndispers.helper import vars2
from itertools import chain

class LBO(Medium):
    """
    LBO (Li B_3 O_5) crystal

    - Point group : 2mm  (C_{2v})
    - Crystal system : orthorhombic 
    - Dielectric principal axes, x // a, y // -c, z // b
    - Biaxial, with two optic axes in xz plane, symmetric with respect to z-axis

    Sellmeier equation
    ------------------
    n(wl) = sqrt(A_i + B_i/(wl**2 - C_i) - D_i * wl**2 + E_i * wl**4)  for i = x,y,z

    Thermo-optic coefficient
    ------------------------
    dn/dT = (F_i + G_i * wl)  for i=x,y,z
    
    Validity range
    --------------


    Ref
    ---
    https://www.newlightphotonics.com/Nonlinear-Optical-Crystals/LBO-Crystals

    Note
    ----
    In the current version, biaxial crystals are limited to the principal dielectric planes, 
    xy, yz or zx planes. In other words, a wavevector of light must be within any one of 
    the three planes. Correspondence between principal plane, polarization orientations of 
    o-wave and e-wave, polar (theta) and azimuthal (phi) angles of a wavevector with respect 
    to z and x principal axes, respectively, are shown in the table below.

    plane  |  o-wave  |  e-wave  |  theta  |   phi   |
    ================================================
    xy     |    z    |    xy   |   pi/2  |   var   |
    yz     |    x    |    yz   |   var   |   pi/2  |
    zx     |    y    |    zx   |   var   |    0    |
    ------------------------------------------------
    pi = 3.14159...
    var : variable

    Example
    -------
    >>> lbo_xy = ndispers.media.crystals.LBO_KK2018_xy()
    >>> lbo_xy.n(0.6, 0.3*pi, 40, pol='e') # for xy plane, 2nd argument is phi_rad. theta_rad is fixed at 0.5*pi.

    """

    __slots__ = ["_A_x", "_B_x", "_C_x", "_D_x", "_E_x",
                 "_A_y", "_B_y", "_C_y", "_D_y", "_E_y",
                 "_A_z", "_B_z", "_C_z", "_D_z", "_E_z",
                 "_F_x", "_F_y", "_F_z", 
                 "_G_x", "_G_y", "_G_z"]
    
    def __init__(self):
        super().__init__()

        # for x-axis
        self._A_x = 2.454140
        self._B_x = 0.011249
        self._C_x = 0.011350
        self._D_x = 0.01459
        self._E_x = -6.60e-5
        # for y-axis
        self._A_y = 2.539070
        self._B_y = 0.012711
        self._C_y = 0.012523
        self._D_y = 0.018540
        self._E_y = +2.00e-4
        # z-axis
        self._A_z = 2.586179
        self._B_z = 0.013099
        self._C_z = 0.011893
        self._D_z = 0.017968
        self._E_z = -2.26e-4
        # dn/dT
        self._F_x = +2.30e-6
        self._F_y = -19.40e-6
        self._F_z = -9.70e-6
        self._G_x = -3.76e-6
        self._G_y = +6.01e-6
        self._G_z = +1.50e-6
    
    @property
    def constants(self):
        print(vars2(self))

    @property
    def symbols(self):
        return [wl, theta, phi, T]

    def _n_T20_x_expr(self):
        """ Sympy expression, dispersion formula for x-axis (principal dielectric axis) at 20degC"""
        return sympy.sqrt(self._A_x + self._B_x/(wl**2 - self._C_x) - self._D_x * wl**2 + self._E_x * wl**4)  
    
    def _n_T20_y_expr(self):
        """ Sympy expression, dispersion formula for y-axis (principal dielectric axis) at 20degC"""
        return sympy.sqrt(self._A_y + self._B_y/(wl**2 - self._C_y) - self._D_y * wl**2 + self._E_y * wl**4)
    
    def _n_T20_z_expr(self):
        """ Sympy expression, dispersion formula for x-axis (principal dielectric axis) at 20degC"""
        return sympy.sqrt(self._A_z + self._B_z/(wl**2 - self._C_z) - self._D_z * wl**2 + self._E_z * wl**4)
    
    def dndT_x_expr(self):
        return self._F_x + self._G_x * wl
    
    def dndT_y_expr(self):
        return self._F_y + self._G_y * wl
    
    def dndT_z_expr(self):
        return self._F_z + self._G_z * wl

    def n_x_expr(self):
        """ sympy expresssion, dispersion formula of x-axis (principal dielectric axis) """
        return self._n_T20_x_expr() + self.dndT_x_expr() * (T - 20)
    
    def n_y_expr(self):
        """ sympy expresssion, dispersion formula of y-axis (principal dielectric axis) """
        return self._n_T20_y_expr() + self.dndT_y_expr() * (T - 20)

    def n_z_expr(self):
        """ sympy expresssion, dispersion formula of z-axis (principal dielectric axis) """
        return self._n_T20_z_expr() + self.dndT_z_expr() * (T - 20)


class LBO_xy(LBO):
    __slots__ = ["_LBO_xy__plane", "_LBO_xy__theta_rad", "_LBO_xy__phi_rad"]

    def __init__(self):
        super().__init__()
        self._LBO_xy__plane = 'xy'
        self._LBO_xy__theta_rad = 0.5*pi
        self._LBO_xy__phi_rad = 'var'
    
    @property
    def help(self):
        print(super().__doc__)

    @property
    def plane(self):
        return self._LBO_xy__plane

    @property
    def theta_rad(self):
        return self._LBO_xy__theta_rad

    @property
    def phi_rad(self):
        return self._LBO_xy__phi_rad

    @property
    def constants(self):
        print({**vars2(super()), **vars2(self)})

    def n_o_expr(self):
        """ sympy expresssion, 
        dispersion formula for o-wave polarization for a given principal plane """
        return super().n_z_expr()
    
    def n_e_expr(self):
        """ sympy expresssion, 
        dispersion formula for e-wave polarization for a given principal plane """
        return super().n_x_expr() * super().n_y_expr() / sympy.sqrt( super().n_x_expr()**2 * sympy.cos(phi)**2 + super().n_y_expr()**2 * sympy.sin(phi)**2 )

    def n_expr(self, pol):
        """ sympy expresssion, 
        dispersion formula for a given polarization """
        if pol == 'o':
            return self.n_o_expr()
        elif pol == 'e':
            return self.n_e_expr()
        else:
            raise ValueError("pol = '%s' must be 'o' or 'e'" % pol)

    def n(self, wl_um, phi_rad, T_degC, pol='o'):
        """
        Refractive index in xy plane.

        input
        ------
        wl_um     :  float or array_like, wavelength in µm
        phi_rad   :  float or array_like, polar angle in radians
        T_degC    :  float or array_like, temperature of crystal in degree C.
        (Note: theta_rad is fixed at 0.5*pi in xy principal plane.)

        return
        -------
        Refractive index, float or array_like
        """
        return super().n(wl_um, 0.5*pi, phi_rad, T_degC, pol=pol)

    def dn_wl(self, wl_um, phi_rad, T_degC, pol='o'):
        return super().dn_wl(wl_um, 0.5*pi, phi_rad, T_degC, pol=pol)
    
    def d2n_wl(self, wl_um, phi_rad, T_degC, pol='o'):
        return super().d2n_wl(wl_um, 0.5*pi, phi_rad, T_degC, pol=pol)

    def d3n_wl(self, wl_um, phi_rad, T_degC, pol='o'):
        return super().d3n_wl(wl_um, 0.5*pi, phi_rad, T_degC, pol=pol)

    def GD(self, wl_um, phi_rad, T_degC, pol='o'):
        """Group Delay [fs/mm]"""
        return super().GD(wl_um, 0.5*pi, phi_rad, T_degC, pol=pol)
    
    def GV(self, wl_um, phi_rad, T_degC, pol='o'):
        """Group Velocity [µm/fs]"""
        return super().GV(wl_um, 0.5*pi, phi_rad, T_degC, pol=pol)
    
    def ng(self, wl_um, phi_rad, T_degC, pol='o'):
        """Group index, c/Group velocity"""
        return super().ng(wl_um, 0.5*pi, phi_rad, T_degC, pol=pol)

    def GVD(self, wl_um, phi_rad, T_degC, pol='o'):
        """Group Delay Dispersion [fs^2/mm]"""
        return super().GVD(wl_um, 0.5*pi, phi_rad, T_degC, pol=pol)

    def TOD(self, wl_um, phi_rad, T_degC, pol='o'):
        """Third Order Dispersion [fs^3/mm]"""
        return super().TOD(wl_um, 0.5*pi, phi_rad, T_degC, pol=pol)
    
    def woa_theta(self, wl_um, phi_rad, T_degC, pol='e'):
        return super().woa_theta(wl_um, 0.5*pi, phi_rad, T_degC, pol=pol)
    
    def woa_phi(self, wl_um, phi_rad, T_degC, pol='e'):
        return super().woa_phi(wl_um, 0.5*pi, phi_rad, T_degC, pol=pol)
    
    def dndT(self, wl_um, phi_rad, T_degC, pol='o'):
        return super().dndT(wl_um, 0.5*pi, phi_rad, T_degC, pol=pol)


class LBO_yz(LBO):
    __slots__ = ["_LBO_yz__plane", "_LBO_yz__theta_rad", "_LBO_yz__phi_rad"]

    def __init__(self):
        super().__init__()
        self._LBO_yz__plane = 'yz'
        self._LBO_yz__phi_rad = 0.5*pi
        self._LBO_yz__theta_rad = 'var'
    
    @property
    def help(self):
        print(super().__doc__)

    @property
    def plane(self):
        return self._LBO_yz__plane

    @property
    def theta_rad(self):
        return self._LBO_yz__theta_rad

    @property
    def phi_rad(self):
        return self._LBO_yz__phi_rad

    @property
    def constants(self):
        print({**vars2(super()), **vars2(self)})

    def n_o_expr(self):
        """ sympy expresssion, 
        dispersion formula for o-wave polarization for yx principal plane """
        return super().n_x_expr()
    
    def n_e_expr(self):
        """ sympy expresssion, 
        dispersion formula for e-wave polarization for yz principal plane """
        return super().n_y_expr() * super().n_z_expr() / sympy.sqrt( super().n_y_expr()**2 * sympy.sin(theta)**2 + super().n_z_expr()**2 * sympy.cos(theta)**2 )

    def n_expr(self, pol):
        """ sympy expresssion, 
        dispersion formula for a given polarization """
        if pol == 'o':
            return self.n_o_expr()
        elif pol == 'e':
            return self.n_e_expr()
        else:
            raise ValueError("pol = '%s' must be 'o' or 'e'" % pol)

    def n(self, wl_um, theta_rad, T_degC, pol='o'):
        """
        Refractive index in yz plane.

        input
        ------
        wl_um     :  float or array_like, wavelength in µm
        theta_rad   :  float or array_like, azimuthal angle in radians
        T_degC    :  float or array_like, temperature of crystal in degree C.
        (Note: phi_rad is fixed at 0.5*pi in xy principal plane.)

        return
        -------
        Refractive index, float or array_like
        """
        return super().n(wl_um, theta_rad, 0.5*pi, T_degC, pol=pol)

    def dn_wl(self, wl_um, theta_rad, T_degC, pol='o'):
        return super().dn_wl(wl_um, theta_rad, 0.5*pi, T_degC, pol=pol)
    
    def d2n_wl(self, wl_um, theta_rad, T_degC, pol='o'):
        return super().d2n_wl(wl_um, theta_rad, 0.5*pi, T_degC, pol=pol)

    def d3n_wl(self, wl_um, theta_rad, T_degC, pol='o'):
        return super().d3n_wl(wl_um, theta_rad, 0.5*pi, pol=pol)

    def GD(self, wl_um, theta_rad, T_degC, pol='o'):
        """Group Delay [fs/mm]"""
        return super().GD(wl_um, theta_rad, 0.5*pi, T_degC, pol=pol)
    
    def GV(self, wl_um, theta_rad, T_degC, pol='o'):
        """Group Velocity [µm/fs]"""
        return super().GV(wl_um, theta_rad, 0.5*pi, T_degC, pol=pol)
    
    def ng(self, wl_um, theta_rad, T_degC, pol='o'):
        """Group index, c/Group velocity"""
        return super().ng(wl_um, theta_rad, 0.5*pi, T_degC, pol=pol)

    def GVD(self, wl_um, theta_rad, T_degC, pol='o'):
        """Group Delay Dispersion [fs^2/mm]"""
        return super().GVD(wl_um, theta_rad, 0.5*pi, T_degC, pol=pol)

    def TOD(self, wl_um, theta_rad, T_degC, pol='o'):
        """Third Order Dispersion [fs^3/mm]"""
        return super().TOD(wl_um, theta_rad, 0.5*pi, T_degC, pol=pol)
    
    def woa_theta(self, wl_um, theta_rad, T_degC, pol='e'):
        return super().woa_theta(wl_um, theta_rad, 0.5*pi, T_degC, pol=pol)
    
    def woa_phi(self, wl_um, theta_rad, T_degC, pol='e'):
        return super().woa_phi(wl_um, theta_rad, 0.5*pi, T_degC, pol=pol)
    
    def dndT(self, wl_um, theta_rad, T_degC, pol='o'):
        return super().dndT(wl_um, theta_rad, 0.5*pi, T_degC, pol=pol)


class LBO_zx(LBO):
    __slots__ = ["_LBO_zx__plane", "_LBO_zx__theta_rad", "_LBO_zx__phi_rad"]

    def __init__(self):
        super().__init__()
        self._LBO_zx__plane = 'zx'
        self._LBO_zx__theta_rad = 'var'
        self._LBO_zx__phi_rad = 0.5*pi
    
    @property
    def help(self):
        print(super().__doc__)

    @property
    def plane(self):
        return self._LBO_zx__plane

    @property
    def theta_rad(self):
        return self._LBO_zx__theta_rad

    @property
    def phi_rad(self):
        return self._LBO_zx__phi_rad

    @property
    def constants(self):
        print({**vars2(super()), **vars2(self)})

    def n_o_expr(self):
        """ sympy expresssion, 
        dispersion formula for o-wave polarization for zx principal plane """
        return super().n_y_expr()
    
    def n_e_expr(self):
        """ sympy expresssion, 
        dispersion formula for e-wave polarization for zx principal plane """
        return super().n_z_expr() * super().n_x_expr() / sympy.sqrt( super().n_z_expr()**2 * sympy.cos(theta)**2 + super().n_x_expr()**2 * sympy.sin(theta)**2 )

    def n_expr(self, pol):
        """ sympy expresssion, 
        dispersion formula for a given polarization """
        if pol == 'o':
            return self.n_o_expr()
        elif pol == 'e':
            return self.n_e_expr()
        else:
            raise ValueError("pol = '%s' must be 'o' or 'e'" % pol)

    def n(self, wl_um, theta_rad, T_degC, pol='o'):
        """
        Refractive index in yz plane.

        input
        ------
        wl_um     :  float or array_like, wavelength in µm
        theta_rad   :  float or array_like, azimuthal angle in radians
        T_degC    :  float or array_like, temperature of crystal in degree C.
        (Note: phi_rad is fixed at 0.5*pi in xy principal plane.)

        return
        -------
        Refractive index, float or array_like
        """
        return super().n(wl_um, theta_rad, 0.5*pi, T_degC, pol=pol)

    def dn_wl(self, wl_um, theta_rad, T_degC, pol='o'):
        return super().dn_wl(wl_um, theta_rad, 0.5*pi, T_degC, pol=pol)
    
    def d2n_wl(self, wl_um, theta_rad, T_degC, pol='o'):
        return super().d2n_wl(wl_um, theta_rad, 0.5*pi, T_degC, pol=pol)

    def d3n_wl(self, wl_um, theta_rad, T_degC, pol='o'):
        return super().d3n_wl(wl_um, theta_rad, 0.5*pi, pol=pol)

    def GD(self, wl_um, theta_rad, T_degC, pol='o'):
        """Group Delay [fs/mm]"""
        return super().GD(wl_um, theta_rad, 0.5*pi, T_degC, pol=pol)
    
    def GV(self, wl_um, theta_rad, T_degC, pol='o'):
        """Group Velocity [µm/fs]"""
        return super().GV(wl_um, theta_rad, 0.5*pi, T_degC, pol=pol)
    
    def ng(self, wl_um, theta_rad, T_degC, pol='o'):
        """Group index, c/Group velocity"""
        return super().ng(wl_um, theta_rad, 0.5*pi, T_degC, pol=pol)

    def GVD(self, wl_um, theta_rad, T_degC, pol='o'):
        """Group Delay Dispersion [fs^2/mm]"""
        return super().GVD(wl_um, theta_rad, 0.5*pi, T_degC, pol=pol)

    def TOD(self, wl_um, theta_rad, T_degC, pol='o'):
        """Third Order Dispersion [fs^3/mm]"""
        return super().TOD(wl_um, theta_rad, 0.5*pi, T_degC, pol=pol)
    
    def woa_theta(self, wl_um, theta_rad, T_degC, pol='e'):
        return super().woa_theta(wl_um, theta_rad, 0.5*pi, T_degC, pol=pol)
    
    def woa_phi(self, wl_um, theta_rad, T_degC, pol='e'):
        return super().woa_phi(wl_um, theta_rad, 0.5*pi, T_degC, pol=pol)
    
    def dndT(self, wl_um, theta_rad, T_degC, pol='o'):
        return super().dndT(wl_um, theta_rad, 0.5*pi, T_degC, pol=pol)



from ndispers._baseclass import Medium
import numpy as np

class Uniax_neg_32(Medium):
    """
    Class of negative uniaxial crystals of point group 32.

    d = [[ d11, -d11,  0, d14,   0,    0  ],
         [  0,    0,   0,  0,  -d14, -d11 ],
         [  0,    0,   0,  0,    0,    0  ]]
      = [d_il]
        i = 1,2,3 (1=x, 2=y, 3=z)
        l = 1,2,...,6 (1=xx, 2=yy, 3=zz, 4=yz, 5=zx, 6=xy)
    
    Under the Kleinman symmetry,
    dxyz = dyzx, or d14 = d25.
    
    With d25 = - d14, we get d14 = d25 = -d14. Therefore, d14 = 0.
    Nonzero element is only d11.

    """

    def __init__(self):
        super().__init__()
    
    def delta11(self, d11_known, wl1, wl2, T_degC):
        """
        Miller' delta for d11 coefficient (ijk = xxx).

        Parameters
        ----------
        d11_known : float, second-order nonlinear coefficient, d11, known from literature.
        wl1 : float, first pump wavelength in µm for the d11
        wl2 : float, second pump wavelength in µm for the d11
        T_degC : float, crystal temperature in degC.

        return
        ------
        float, Miller's delta for 11 (ijk = xxx) component.

        """
        wl3 = 1./(1./wl1 + 1./wl2)
        theta_rad = 0 # This is an arbitrary value since n of o-wave does not depends on theta.
        # chi : linear susceptibility
        chi_wl1o = self.n(wl1, theta_rad, T_degC, pol='o')**2 - 1
        chi_wl2o = self.n(wl2, theta_rad, T_degC, pol='o')**2 - 1
        chi_wl3o = self.n(wl3, theta_rad, T_degC, pol='o')**2 - 1
        return d11_known / (chi_wl3o * chi_wl2o * chi_wl1o)

    def d11_sfg(self, wl1, wl2, T_degC, delta11=0):
        """
        d11 coefficient as a function of two pump wavelengths for SFG.

        Parameters
        ----------
        wl1 : float, first pump wavelength in µm.
        wl2 : float, second pump wavelength in µm.
        T_degC : crystal temperature in degC.

        return
        ------
        float, d11 coefficient (pm/V).

        """

        wl3 = 1./(1./wl1 + 1./wl2)
        theta_rad = 0 # This is an arbitrary value since n of o-wave does not depends on theta.
        chi_wl1o = self.n(wl1, theta_rad, T_degC, pol='o')**2 - 1
        chi_wl2o = self.n(wl2, theta_rad, T_degC, pol='o')**2 - 1
        chi_wl3o = self.n(wl3, theta_rad, T_degC, pol='o')**2 - 1
        return delta11 * chi_wl3o * chi_wl2o * chi_wl1o
    
    def delta14(self, d14_known, wl1, wl2, T_degC):
        """
        Miller' delta for d14 coefficient (ijk = xyz).

        Parameters
        ----------
        d14_known : float, second-order nonlinear coefficient, d14, known from literature.
        wl1 : float, first pump wavelength in µm for the d14
        wl2 : float, second pump wavelength in µm for the d14
        T_degC : float, crystal temperature in degC.

        return
        ------
        float, Miller's delta for 14 (ijk = xyz) component.

        """
        wl3 = 1./(1./wl1 + 1./wl2)
        theta_rad = 0 # This is an arbitrary value since n of o-wave does not depends on theta.
        # chi : linear susceptibility
        chi_wl1o = self.n(wl1, theta_rad, T_degC, pol='o')**2 - 1
        chi_wl2o = self.n(wl2, theta_rad, T_degC, pol='o')**2 - 1
        chi_wl3o = self.n(wl3, theta_rad, T_degC, pol='o')**2 - 1
        return d14_known / (chi_wl3o * chi_wl2o * chi_wl1o)

    def d14_sfg(self, wl1, wl2, T_degC, delta14=0):
        """
        d11 coefficient as a function of two pump wavelengths for SFG.

        Parameters
        ----------
        wl1 : float, first pump wavelength in µm.
        wl2 : float, second pump wavelength in µm.
        T_degC : crystal temperature in degC.

        return
        ------
        float, d14 coefficient (pm/V).

        """

        wl3 = 1./(1./wl1 + 1./wl2)
        theta_rad = 0 # This is an arbitrary value since n of o-wave does not depends on theta.
        chi_wl1e = self.n(wl1, theta_rad, T_degC, pol='e')**2 - 1
        chi_wl2o = self.n(wl2, theta_rad, T_degC, pol='o')**2 - 1
        chi_wl3e = self.n(wl3, theta_rad, T_degC, pol='e')**2 - 1
        return delta14 * chi_wl3e * chi_wl2o * chi_wl1e
    
    def deff_sfg(self, wl1, wl2, theta_rad, phi_rad, T_degC, pol1, pol2, pol3):
        """
        Effective second-order nonlinear coefficient, d_eff, of SFG.
        Wavelength dependence is estimated from Miller's rule.

        Parameters
        ----------
        wl1  :  float, first pump wavelength in µm.
        wl2  : float, second pump wavelength in µm.
        T_degC : float, crystal temperature in degC.
        pol1 : str, {'o', 'e'}, polarization of the 1st pump wave.
        pol2 : str, {'o', 'e'}, polarization of the 2nd pump wave.
        pol3 : str, {'o', 'e'}, polarization of the sum-frequency wave.

        return
        ------
        float, Effective d coefficient (pm/V).

        """
        wl3 = 1./(1./wl1 + 1./wl2)
        _d11 = self.d11_sfg(wl1, wl2, T_degC)
        _d14 = self.d14_sfg(wl1, wl2, T_degC)
        rho3 = self.woa_theta(wl3, theta_rad, T_degC, pol=pol3)
        # For negative uniaxial type-I (ooe),
        if pol1 == 'o' and pol2 == 'o' and pol3 == 'e':
            deff = _d11 * np.cos(3*phi_rad) * np.cos(theta_rad + rho3)
        
        # For negative uniaxial type-II (oee),
        elif pol1 == 'o' and pol2 == 'e' and pol3 == 'e':
            rho2 = self.woa_theta(wl2, theta_rad, T_degC, pol=pol2)
            deff = _d11 * np.sin(3*phi_rad) * np.cos(theta_rad + rho2) * np.cos(theta_rad + rho3)\
                + _d14 * np.sin(theta_rad + rho2) * np.cos(theta_rad + rho3)
        
        # For negative uniaxial type-II (eoe),
        elif pol1 == 'e' and pol2 == 'o' and pol3 == 'e':
            rho1 = self.woa_theta(wl1, theta_rad, T_degC, pol=pol1)
            deff = _d11 * np.sin(3*phi_rad) * np.cos(theta_rad + rho1) * np.cos(theta_rad + rho3)\
                + _d14 * np.sin(theta_rad + rho1) * np.cos(theta_rad + rho3)
        
        else:
            raise ValueError("For negative uniaxial crystals,pol1, pol2, pol3 must be among {ooe, oee, eoe}.")

        return deff



"""
ndispers.media.crystals/__init__.py
anisotropic crystals
"""

from ._alphaBBO import AlphaBBO
from ._betaBBO_Eimerl1987 import BetaBBO as BetaBBO_Eimerl1987
from ._betaBBO_Ghosh1995 import BetaBBO as BetaBBO_Ghosh1995
from ._betaBBO_KK2010 import BetaBBO as BetaBBO_KK2010
from ._betaBBO_Tamosauskas2018 import BetaBBO as BetaBBO_Tamosauskas2018
from ._calcite import Calcite
from ._CLBO import CLBO
from ._KBBF_Li2016 import KBBF
from ._KDP import KDP
from ._KTP import KTP_xy, KTP_yz, KTP_zx
from ._LB4 import LB4
from ._LBO_Castech import LBO_xy as LBO_Castech_xy
from ._LBO_Castech import LBO_yz as LBO_Castech_yz
from ._LBO_Castech import LBO_zx as LBO_Castech_zx
from ._LBO_Ghosh1995 import LBO_xy as LBO_Ghosh1995_xy
from ._LBO_Ghosh1995 import LBO_yz as LBO_Ghosh1995_yz
from ._LBO_Ghosh1995 import LBO_zx as LBO_Ghosh1995_zx
from ._LBO_KK1994 import LBO_xy as LBO_KK1994_xy
from ._LBO_KK1994 import LBO_yz as LBO_KK1994_yz
from ._LBO_KK1994 import LBO_zx as LBO_KK1994_zx
from ._LBO_KK2018 import LBO_xy as LBO_KK2018_xy
from ._LBO_KK2018 import LBO_yz as LBO_KK2018_yz
from ._LBO_KK2018 import LBO_zx as LBO_KK2018_zx
from ._LBO_Newlight import LBO_xy as LBO_Newlight_xy
from ._LBO_Newlight import LBO_yz as LBO_Newlight_yz
from ._LBO_Newlight import LBO_zx as LBO_Newlight_zx
from ._RBBF_Chen2009 import RBBF
from ._SLN_MgO_doped import SLN
from ._SLT_MgO_doped import SLT